//: arithm.h : Ganzzzahlarithmetik mit Zeichenketten - R.Richter 2004-10-01
///////////////////////////////////////////////////////////////////////////

#ifndef ARITHM_H
#define ARITHM_H
#include <string>
// #include <iostream>  // temporr

typedef std::string Zahl;

Zahl add(Zahl a, Zahl b);
Zahl sub(Zahl a, Zahl b);
Zahl mul(Zahl a, Zahl b);
Zahl div(Zahl a, Zahl b);
Zahl mod(Zahl a, Zahl b);

#endif // ARITHM_H

//: arithm.cpp : Ganzzzahlarithmetik mit Zeichenketten - R.Richter 2004-10-01
/////////////////////////////////////////////////////////////////////////////
// Hilfsfunktionen
#include <cassert>
#include <algorithm>
// #include "arithm.h"
using namespace std;

namespace {
// bei MSVC : keine Funktionen min, max
#ifdef _MSC_VER
template<class T>
const T& min(const T& x, const T& y)
{
  return x<y? x:y;
}

template<class T>
const T& max(const T& x, const T& y)
{
  return x<y? y:x;
}
#endif // _MSC_VER
int ziffer(char c)
{
  assert('0' <= c && c <= '9');
  return c - '0';
}

char zeichen(int ziffer)
{
  assert(0 <= ziffer && ziffer < 10);
  return '0' + ziffer;
}

bool ist_kleiner(Zahl abs_a, Zahl abs_b)
{
  assert(abs_a.find_first_not_of("0123456789") == std::string::npos);
  assert(abs_b.find_first_not_of("0123456789") == std::string::npos);
    
  if(abs_a.size() <  abs_b.size()) return true;
  if(abs_a.size() == abs_b.size() && abs_a < abs_b) return true;
  return false;   
}    

bool istnegativ(Zahl n)
{
  assert(n.size() > 0);  
  return n[0] == '-';  
}    

Zahl absolutwert(Zahl n)
{
  if(istnegativ(n)) return n.substr(1);
  return n;
}

Zahl entgegengesetztes(Zahl n)
{
  if(n == "0") return n;   
  return istnegativ(n)? absolutwert(n) : "-" + n;  
}    

void umkehren(std::string& s)
{
  std::reverse(s.begin(), s.end());
}

void entferne_nullen(std::string& s)  // am Ende 
{
  while(s.size()>1 && *(s.end()-1) == '0')
  {
    s.erase(s.end()-1);  
  }
}    

void abs_addition(Zahl& sum, Zahl a, Zahl b)
{
  assert(a.size() > 0);  
  assert(b.size() > 0); 
  umkehren(a);
  umkehren(b); 

  int m = a.size();
  int n = b.size();
  int uebertrag = 0;

  for(int i=0; i < max(m, n); ++i)
  {
    if(i < m) uebertrag += ziffer(a[i]);
    if(i < n) uebertrag += ziffer(b[i]);
    sum       += zeichen(uebertrag%10);
    uebertrag /= 10;
  }
  if(uebertrag) sum += zeichen(uebertrag);

  umkehren(sum);    
}

void abs_subtraktion(Zahl& dif, Zahl a, Zahl b)
{
  assert(a.size() > 0);  
  assert(b.size() > 0);  
  umkehren(a);
  umkehren(b); 

  int m = a.size();
  int n = b.size();
  int uebertrag = 1; // geborgt

  for(int i=0; i < max(m, n); ++i)
  {
    uebertrag += (10-1); // borgen
    if(i < m) uebertrag += ziffer(a[i]);
    if(i < n) uebertrag -= ziffer(b[i]);
    dif       += zeichen(uebertrag%10);
    uebertrag /= 10;
  }
  assert(uebertrag==1); // muss brig bleiben, sonst war a<b

  entferne_nullen(dif);
  umkehren(dif); 
}

void abs_multiplikation(Zahl& prod, Zahl a, Zahl b)
{
  assert(a.size() > 0);  
  assert(b.size() > 0);  
  umkehren(a);  umkehren(b); 

  int m = a.size();
  int n = b.size();
  prod = std::string(m + n, '0');
  
  for(int j = 0; j<n; ++j)
  {
    if(ziffer(b[j]) != 0)
    {    
      int uebertrag = 0; 
      for(int i = 0; i<m; ++i)
      {
        uebertrag += ziffer(prod[i+j]);
        uebertrag += ziffer(a[i]) * ziffer(b[j]);
        prod[i+j]  = zeichen(uebertrag%10);
        uebertrag /= 10;
      }
      prod[m+j] = zeichen(uebertrag);    
    }    
  }  

  entferne_nullen(prod);
  umkehren(prod);
}

void einstellige_division(Zahl& u, int teiler, int& rest)
{
  // Absolutzahl u[] in umgekehrter Folge  
  int i = u.size();
  int uebertrag = 0;
  while(i-- > 0)
  {
    uebertrag *= 10;
    uebertrag += ziffer(u[i]);
    u[i]       = zeichen(uebertrag/teiler);
    uebertrag %= teiler;      
  }    
  rest = uebertrag;
}  

void einstellige_multiplikation(Zahl& u, int faktor)
{
  // Absolutzahl u[] in umgekehrter Folge  
  int m = u.size();
  int uebertrag = 0;
  
  for(int i=0; i<m; ++i)  
  {
    uebertrag += ziffer(u[i]) * faktor;
    u[i]       = zeichen(uebertrag%10);
    uebertrag /= 10;    
  }      
  if(uebertrag)
  {
    u += zeichen(uebertrag);
  }    
}  

// bool debugpause = false;
// int nachkorrekturen = 0;
// int rueckadditionen = 0;

void abs_division(Zahl& quot, Zahl& rest, Zahl a, Zahl b)
{
  assert(a.size() > 0);  
  assert(b.size() > 0);  
  assert(b != "0");  

  if(ist_kleiner(a, b))
  {
    quot = "0";
    rest = a;
    return;   
  }    
  umkehren(a);
  umkehren(b); 

  if(b.size() == 1)
  {
    quot = a;
    int r = 0;
    einstellige_division(quot, ziffer(b[0]), r);  
    rest = "";
    rest += zeichen(r);
  }        
  else // mehrstellige Division
  {
    // Lit: DEK: TAOCP. Vol. 2, pp. 270--275
    int n = b.size();
    int m = a.size() - n;
    Zahl& u = rest;
    Zahl& v = b;
    u = a;
    u.resize(m + n + 1, '0');
    quot.resize(m + 1, '0');

    // normalisieren: sichern, dass hchste Ziffer von v >= 10/2
    int scale = 10 / (1 + ziffer(v[n-1]));
    if(scale > 1)
    {
      einstellige_multiplikation(u, scale);
      einstellige_multiplikation(v, scale);
    }    
    
    // dividieren
    // erste zwei ziffern des Divisors b    
    int v1 = ziffer(v[n-1]);
    int v2 = ziffer(v[n-2]);
    
    for(int j = m; j>=0; --j)
    {
      // schtze nchste Ziffer qhat des Quotienten  
      int u12 = ziffer(u[j+n])*10 + ziffer(u[j+n-1]);
      int qhat = u12/v1;        
      int rhat = u12%v1; 

      // cout << u << endl;
      // cout << v << endl;
      // cout << "1. Naeherung q[" << j << "]=" << qhat << endl;

      if(qhat == 10 || qhat*v2 > 10*rhat + ziffer(u[j+n-2]))
      {
        qhat--;
        rhat += v1;

        // cout << "1. Nachkorrektur " << qhat << endl;

        if(rhat < 10 && qhat*v2 > 10*rhat + ziffer(u[j+n-2]))
        {
            qhat--;
            // cout << "2. Nachkorrektur " << qhat << endl;
            // nachkorrekturen++;
            // if(debugpause) system("Pause");
        }  
      }
      // nun ist q[j] fast immer korrekt 
      // multipliziere und ziehe ab
      int uebertrag = 1; // borgen
      int prod = 0;
      for(int i=0; i<n; ++i)
      {
        prod      += qhat*ziffer(v[i]);
        uebertrag += (10-1) + ziffer(u[j+i]) - prod%10;
        u[j+i]     = zeichen(uebertrag%10);
        uebertrag /= 10;
        prod      /= 10;   
      }    
      // vorderste Ziffer
      uebertrag += (10-1) + ziffer(u[j+n]) - prod%10;
      u[j+n]     = zeichen(uebertrag%10);
      uebertrag /= 10;
      
      if(uebertrag == 0) // doch einmal zuviel weggenommen, zurckaddieren
      {  
        qhat--;
        // cout << "Rueckaddition qhat=" << qhat << endl;
        // rueckadditionen++;
        // if(debugpause) system("Pause");
        for(int i=0; i<n; ++i)
        {
          uebertrag += ziffer(u[j+i]);
          uebertrag += ziffer(v[i]);
          u[j+i]     = zeichen(uebertrag%10);  
          uebertrag /= 10;            
        }
        u[j+n] = zeichen(0);    
       }    
      // Ziffer aufschreiben 
      quot[j] = zeichen(qhat);
    }    

    // denormalisieren: rest durch d
    if(scale > 1)
    {
      int r = 0;
      einstellige_division(rest, scale, r);  
      assert(r == 0);
    }    
  }
  entferne_nullen(quot);
  entferne_nullen(rest);
  umkehren(quot);
  umkehren(rest); 
}

} // namespace

// Implementierung
Zahl add(Zahl a, Zahl b)
{
  if(istnegativ(a) == istnegativ(b))
  {  
    Zahl sum;
    abs_addition(sum, absolutwert(a), absolutwert(b));
    return istnegativ(a)? entgegengesetztes(sum) : sum;
  }
  Zahl dif;
  Zahl abs_a = absolutwert(a);
  Zahl abs_b = absolutwert(b);
  
  if(ist_kleiner(abs_b, abs_a))
  {
    abs_subtraktion(dif, abs_a, abs_b);   
    if(istnegativ(a)) 
    { 
      dif = entgegengesetztes(dif);
    }    
  }
  else
  {
    abs_subtraktion(dif, abs_b, abs_a);   
    if(istnegativ(b)) 
    {
      dif = entgegengesetztes(dif);
    }    
  }
  return dif;
}

Zahl sub(Zahl a, Zahl b)
{
  return add(a, entgegengesetztes(b));
}

Zahl mul(Zahl a, Zahl b)
{
  Zahl prod;
  abs_multiplikation(prod, absolutwert(a), absolutwert(b));  
  return istnegativ(a) == istnegativ(b) ? prod : entgegengesetztes(prod);  
}

/// Ganzzahlige Division
/// a : b = q Rest r mit q = a div b und r = a mod b 

/// Ergebnisse erfllen Bedingung
/// a = q*b + r

/// Vorzeichenregeln
/// | a | q | b | r
/// | + | + | + | +
/// | + | - | - | +
/// | - | + | - | -
/// | - | - | + | -

Zahl div(Zahl a, Zahl b)
{
  Zahl quot, rest;
  abs_division(quot, rest, absolutwert(a), absolutwert(b));
  return istnegativ(a) == istnegativ(b) ? quot : entgegengesetztes(quot);  
}

Zahl mod(Zahl a, Zahl b)
{
  Zahl quot, rest;
  abs_division(quot, rest, absolutwert(a), absolutwert(b));
  return !istnegativ(a) ? rest : entgegengesetztes(rest);  
}

//: tstarith.cpp : Testprogramm Ganzzahlarithmetik - R.Richter 2004-10-03
/////////////////////////////////////////////////////////////////////////

#include <iostream>
#include <cstdlib> // rand(), srand()
#include <cstdio>  // sprintf()
#include <ctime>   // time()
// #include "arithm.h"
using namespace std;

string zufall()
{
  char buf[20];
  int x = rand();
  if( RAND_MAX < 100000 ) // schlechter Zufallsgenerator  
  {
    x *= rand();
  }
  sprintf(buf, "%d", x);
  return buf;
}

void divtest(int anzahl)
{    
  int seed(time(NULL));
  srand(seed);
  cout << "Startwert=" << seed << endl; 
  cout << "maximale Zufallszahl = " << RAND_MAX << endl;

  for(int i=0; i<anzahl; ++i)
  {  
    string a_strg = zufall();
    if(a_strg != "0")
    {
      a_strg += zufall() + zufall() + zufall(); // bis zu 40 Stellen
    }    
    string b_strg = zufall(); 
    while(b_strg == "0")
    {
      b_strg = zufall();   
    }
    b_strg += zufall(); // bis zu 20 Stellen   
   
    Zahl a = a_strg;
    Zahl b = b_strg; 
    Zahl q = div(a, b);
    Zahl r = mod(a, b);
//    cout << a << "/" << b << "=" << q << endl;                
//    cout << a << "%" << b << "=" << r << endl;   
    Zahl probe = add(mul(q,b), r);
//    cout << "Probe: " << a << " == " << probe << " ? ";
    if(a != probe)
    {
      cout << a << "/" << b << "=" << q << endl;                
      cout << a << "%" << b << "=" << r << endl;   
      cout << "Probe: " << a << " == " << probe << " ? ";
      cout << "Fehler bei Division! Programmabbruch." << endl;
      system("pause");
      exit(1);
    }
    else
    {
//      cout << "ok." << endl;
    }            
  }    
  
  cout << anzahl << " Tests bestanden." << endl; 
//  cout << nachkorrekturen << " Nachkorrekturen" << endl;
//  cout << rueckadditionen << " Rueckadditionen" << endl; 
  cout << time(0) - seed << " Sekunden." << endl;
}    

int main()
{
/*
//  Zahl a = "-9876543210", b = "-12345678";
  Zahl a = "8888888881", b = "99";
//  Zahl a = "950", b = "23";
//  Zahl a = "950", b = "23";

  Zahl s = add(a, b);
  Zahl d = sub(a, b);
  Zahl p = mul(a, b);
  Zahl q = div(a, b);
  Zahl r = mod(a, b);
  
  cout << a << "+" << b << "=" << s << endl;                
  cout << a << "-" << b << "=" << d << endl;                
  cout << a << "*" << b << "=" << p << endl;                
  cout << a << "/" << b << "=" << q << endl;                
  cout << a << "%" << b << "=" << r << endl;   
*/
  divtest(1000000);                          
  cin.get();  
  return 0;   
}    
